@@ -1,13 +1,13 @@
 ;;; embark-consult.el --- Consult integration for Embark -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2021, 2022  Free Software Foundation, Inc.
+;; Copyright (C) 2021-2025  Free Software Foundation, Inc.
 
 ;; Author: Omar Antolín Camarena <omar@matem.unam.mx>
 ;; Maintainer: Omar Antolín Camarena <omar@matem.unam.mx>
 ;; Keywords: convenience
-;; Version: 0.7
-;; Homepage: https://github.com/oantolin/embark
-;; Package-Requires: ((emacs "27.1") (embark "0.20") (consult "0.17"))
+;; Version: 1.1
+;; URL: https://github.com/oantolin/embark
+;; Package-Requires: ((emacs "28.1") (compat "30") (embark "1.1") (consult "1.8"))
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -103,11 +103,13 @@
 (setf (alist-get 'consult-location embark-default-action-overrides)
       #'embark-consult-goto-location)
 
-(defun embark-consult-export-occur (lines)
+(defun embark-consult-export-location-occur (lines)
   "Create an occur mode buffer listing LINES.
-The elements of LINES are assumed to be values of category `consult-line'."
+The elements of LINES should be completion candidates with
+category `consult-line'."
   (let ((buf (generate-new-buffer "*Embark Export Occur*"))
         (mouse-msg "mouse-2: go to this occurrence")
+        (inhibit-read-only t)
         last-buf)
     (with-current-buffer buf
       (dolist (line lines)
@@ -115,32 +117,80 @@ The elements of LINES are assumed to be values of category `consult-line'."
             ((`(,loc . ,num) (consult--get-location line))
              ;; the text properties added to the following strings are
              ;; taken from occur-engine
-             (lineno (propertize (format "%7d:" num)
-                                 'occur-prefix t
-                                 ;; Allow insertion of text at the end
-                                 ;; of the prefix (for Occur Edit mode).
-                                 'front-sticky t
-                                 'rear-nonsticky t
-                                 'occur-target loc
-                                 'follow-link t
-                                 'help-echo mouse-msg))
+             (lineno (propertize
+                      (format "%7d:" num)
+                      'occur-prefix t
+                      ;; Allow insertion of text at the end
+                      ;; of the prefix (for Occur Edit mode).
+                      'front-sticky t
+                      'rear-nonsticky t
+                      'read-only t
+                      'occur-target loc
+                      'follow-link t
+                      'help-echo mouse-msg
+                      'font-lock-face list-matching-lines-prefix-face
+                      'mouse-face 'highlight))
              (contents (propertize (embark-consult--strip line)
                                    'occur-target loc
                                    'occur-match t
                                    'follow-link t
-                                   'help-echo mouse-msg))
+                                   'help-echo mouse-msg
+                                   'mouse-face 'highlight))
              (nl (propertize "\n" 'occur-target loc))
              (this-buf (marker-buffer loc)))
           (unless (eq this-buf last-buf)
             (insert (propertize
                      (format "lines from buffer: %s\n" this-buf)
-                     'face list-matching-lines-buffer-name-face))
+                     'face list-matching-lines-buffer-name-face
+                     'read-only t))
             (setq last-buf this-buf))
-          (insert (concat lineno contents nl))))
+          (insert lineno contents nl)))
       (goto-char (point-min))
       (occur-mode))
     (pop-to-buffer buf)))
 
+(defun embark-consult-export-location-grep (lines)
+  "Create a grep mode buffer listing LINES.
+Any LINES that come from a buffer which is not visiting a file
+will be excluded from the grep buffer, since grep mode only works
+with files.  The elements of LINES should be completion
+candidates with category `consult-location'.  No matches will be
+highlighted in the exported buffer, since the `consult-location'
+candidates do not carry that information."
+  (let (non-file-buffers)
+    (embark-consult--export-grep
+     :header "Exported line search results (file-backed buffers only):\n\n"
+     :lines lines
+     :insert
+     (lambda (lines)
+       (let ((count 0))
+         (dolist (line lines)
+           (pcase-let* ((`(,loc . ,num) (consult--get-location line))
+                        (lineno (format "%d" num))
+                        (contents (embark-consult--strip line))
+                        (buffer (marker-buffer loc))
+                        (file (buffer-file-name buffer)))
+             (if (null file)
+                 (cl-pushnew buffer non-file-buffers)
+               (insert (file-relative-name file) ":" lineno ":" contents "\n")
+               (cl-incf count))))
+         count))
+     :footer
+     (lambda ()
+       (when non-file-buffers
+         (let ((start (goto-char (point-max))))
+           (insert "\nSome results were in buffers with no associated file"
+                   " and are missing\nfrom the exported result:\n")
+           (dolist (buf non-file-buffers)
+             (insert "- " (buffer-name buf) "\n"))
+           (insert "\nEither save the buffers or use the"
+                   " `embark-consult-export-location-occur'\nexporter.")
+           (message "This exporter does not support non-file buffers: %s"
+                    non-file-buffers)
+           (add-text-properties
+            start (point-max)
+            '(read-only t wgrep-footer t front-sticky t))))))))
+
 (defun embark-consult--upgrade-markers ()
   "Upgrade consult-location cheap markers to real markers.
 This function is meant to be added to `embark-collect-mode-hook'."
@@ -149,67 +199,93 @@ This function is meant to be added to `embark-collect-mode-hook'."
       (when (car entry)
         (consult--get-location (car entry))))))
 
+;; Set default `occur-mode' based exporter for consult-line,
+;; consult-line-multi, consult-outline and alike Another option is
+;; using grep-mode by using `embark-consult-export-location-grep'
 (setf (alist-get 'consult-location embark-exporters-alist)
-      #'embark-consult-export-occur)
+      #'embark-consult-export-location-occur)
 (cl-pushnew #'embark-consult--upgrade-markers embark-collect-mode-hook)
 
 ;;; Support for consult-grep
 
 (defvar grep-mode-line-matches)
 (defvar grep-num-matches-found)
-(defvar wgrep-header/footer-parser)
-(declare-function wgrep-setup "ext:wgrep")
+(declare-function compilation--ensure-parse "compile")
 
-(defvar-keymap embark-consult-revert-map
-  :doc "A keymap with a binding for revert-buffer."
+(defvar-keymap embark-consult-rerun-map
+  :doc "A keymap with a binding for `embark-rerun-collect-or-export'."
   :parent nil
-  "g" #'revert-buffer)
+  "g" #'embark-rerun-collect-or-export)
 
-(defun embark-consult-export-grep (lines)
-  "Create a grep mode buffer listing LINES."
-  (let ((buf (generate-new-buffer "*Embark Export Grep*"))
-        (count 0)
-        prop)
+(cl-defun embark-consult--export-grep (&key header lines insert footer)
+  "Create a grep mode buffer listing LINES.
+The HEADER string is inserted at the top of the buffer.  The
+function INSERT is called to insert the LINES and should return a
+count of the matches (there may be more than one match per line).
+The function FOOTER is called to insert a footer."
+  (let ((buf (generate-new-buffer "*Embark Export Grep*")))
     (with-current-buffer buf
-      (insert (propertize "Exported grep results:\n\n" 'wgrep-header t))
-      (dolist (line lines) (insert line "\n"))
-      (goto-char (point-min))
-      (while (setq prop (text-property-search-forward
-                         'face 'consult-highlight-match t))
-        (cl-incf count)
-        (put-text-property (prop-match-beginning prop)
-                           (prop-match-end prop)
-                           'font-lock-face
-                           'match))
-      (goto-char (point-min))
       (grep-mode)
-      (when (> count 0)
-        (setq-local grep-num-matches-found count
-                    mode-line-process grep-mode-line-matches))
+      (setq-local mode-line-process grep-mode-line-matches)
+      (let ((inhibit-read-only t))
+        (insert (propertize header 'wgrep-header t 'front-sticky t))
+        (dlet ((compilation-filter-start (point)))
+          (setq-local grep-num-matches-found (funcall insert lines))
+          (goto-char (point-max))
+          ;; Ensure that all `compilation-message' text properties are added.
+          (compilation--ensure-parse (point))
+          ;; Instead of (run-hooks 'compilation-filter-hook), we only run the
+          ;; Emacs 30 grep heading filter. The other `compilation-filter-hook'
+          ;; functions handle escape sequences, which we do not need here.
+          (when (and (bound-and-true-p grep-use-headings)
+                     (fboundp 'grep--heading-filter))
+            (grep--heading-filter))
+          (funcall footer)
+          (goto-char (point-min))))
       ;; Make this buffer current for next/previous-error
       (setq next-error-last-buffer buf)
       ;; Set up keymap before possible wgrep-setup, so that wgrep
       ;; restores our binding too when the user finishes editing.
       (use-local-map (make-composed-keymap
-                      embark-consult-revert-map
+                      embark-consult-rerun-map
                       (current-local-map)))
-      (setq-local wgrep-header/footer-parser #'ignore)
-      (when (fboundp 'wgrep-setup) (wgrep-setup)))
+      ;; NOTE Wgrep is not needed anymore on Emacs 31 with `grep-edit-mode'.
+      (when (fboundp 'wgrep-setup)
+        ;; TODO Wgrep 3.0 and development versions use different names for the
+        ;; parser variable.
+        (defvar wgrep-header/footer-parser)
+        (defvar wgrep-header&footer-parser)
+        (setq-local wgrep-header/footer-parser #'ignore
+                    wgrep-header&footer-parser #'ignore)
+        (wgrep-setup)))
     (pop-to-buffer buf)))
 
+(defun embark-consult-export-grep (lines)
+  "Create a grep mode buffer listing LINES.
+The elements of LINES should be completion candidates with
+category `consult-grep'."
+  (embark-consult--export-grep
+   :header "Exported grep results:\n\n"
+   :lines lines
+   :insert
+   (lambda (lines)
+     (dolist (line lines) (insert line "\n"))
+     (goto-char (point-min))
+     (let ((count 0) prop)
+       (while (setq prop (text-property-search-forward
+                          'face 'consult-highlight-match t))
+         (cl-incf count)
+         (put-text-property (prop-match-beginning prop)
+                            (prop-match-end prop)
+                            'font-lock-face
+                            'match))
+       count))
+   :footer #'ignore))
+
 (defun embark-consult-goto-grep (location)
   "Go to LOCATION, which should be a string with a grep match."
-  ;; Actions are run in the target window, so in this case whatever
-  ;; window was selected when the command that produced the
-  ;; xref-location candidates ran.  In particular, we inherit the
-  ;; default-directory of the buffer in that window, but we really
-  ;; want the default-directory of the minibuffer or collect window we
-  ;; call the action from, which is the previous window, since the
-  ;; location is given relative to that directory.
-  (let ((default-directory (with-selected-window (previous-window)
-                             default-directory)))
-    (consult--jump (consult--grep-position location))
-    (pulse-momentary-highlight-one-line (point))))
+  (consult--jump (consult--grep-position location))
+  (pulse-momentary-highlight-one-line (point)))
 
 (setf (alist-get 'consult-grep embark-default-action-overrides)
       #'embark-consult-goto-grep)
@@ -218,8 +294,9 @@ This function is meant to be added to `embark-collect-mode-hook'."
 
 ;;; Support for consult-xref
 
-(declare-function xref--show-xref-buffer "ext:xref")
 (declare-function consult-xref "ext:consult-xref")
+(declare-function xref--show-xref-buffer "xref")
+(declare-function xref-pop-to-location "xref")
 (defvar xref-auto-jump-to-first-xref)
 (defvar consult-xref--fetcher)
 
@@ -244,12 +321,12 @@ This function is meant to be added to `embark-collect-mode-hook'."
                        'minibuffer-exit-hook
                        (lambda ()
                          (throw 'xref-items
-                           (xref-items
-                            (or
-                             (plist-get
-                              (embark--maybe-transform-candidates)
-                              :candidates)
-                             (user-error "No candidates for export")))))
+                                (xref-items
+                                 (or
+                                  (plist-get
+                                   (embark--maybe-transform-candidates)
+                                   :candidates)
+                                  (user-error "No candidates for export")))))
                        nil t))
                   (consult-xref fetcher))))))
         `((fetched-xrefs . ,(xref-items items))
@@ -260,17 +337,26 @@ This function is meant to be added to `embark-collect-mode-hook'."
 (setf (alist-get 'consult-xref embark-exporters-alist)
       #'embark-consult-export-xref)
 
+(defun embark-consult-xref (cand)
+  "Default action override for `consult-xref', open CAND xref location."
+  (xref-pop-to-location (get-text-property 0 'consult-xref cand)))
+
+(setf (alist-get 'consult-xref embark-default-action-overrides)
+      #'embark-consult-xref)
+
 ;;; Support for consult-find and consult-locate
 
-(setf (alist-get '(file . consult-find) embark-default-action-overrides)
+(setf (alist-get '(file . consult-find) embark-default-action-overrides
+                 nil nil #'equal)
       #'find-file)
 
-(setf (alist-get '(file . consult-locate) embark-default-action-overrides)
+(setf (alist-get '(file . consult-locate) embark-default-action-overrides
+                 nil nil #'equal)
       #'find-file)
 
-;;; Support for consult-isearch
+;;; Support for consult-isearch-history
 
-(setf (alist-get 'consult-isearch embark-transformer-alist)
+(setf (alist-get 'consult-isearch-history embark-transformer-alist)
       #'embark-consult--target-strip)
 
 ;;; Support for consult-man and consult-info
@@ -336,6 +422,8 @@ This function is meant to be added to `embark-collect-mode-hook'."
 
 (map-keymap
  (lambda (_key cmd)
+   (cl-pushnew 'embark--unmark-target
+               (alist-get cmd embark-pre-action-hooks))
    (cl-pushnew 'embark--allow-edit
                (alist-get cmd embark-target-injection-hooks)))
  embark-consult-search-map)
@@ -356,34 +444,26 @@ This is intended to be used in `embark-target-injection-hooks'."
   (cl-pushnew #'embark-consult--unique-match
               (alist-get cmd embark-target-injection-hooks)))
 
-(cl-defun embark-consult--prep-async (&key type target &allow-other-keys)
-  "Either add Consult's async separator or ignore the TARGET depending on TYPE.
-If the TARGET of the given TYPE has an associated notion of
-directory, we don't want to search for the text of target, but
-rather just start a search in the associated directory.
-
-This is intended to be used in `embark-target-injection-hooks'
-for any action that is a Consult async command."
-  (let* ((style (alist-get consult-async-split-style
-                           consult-async-split-styles-alist))
-         (initial (plist-get style :initial))
-         (separator (plist-get style :separator))
-         (directory (embark--associated-directory target type)))
-    (when directory
-      (delete-minibuffer-contents))
-    (when initial
-      (goto-char (minibuffer-prompt-end))
-      (insert initial)
-      (goto-char (point-max)))
-    (when (and separator (null directory))
-      (goto-char (point-max))
-      (insert separator))))
+(cl-defun embark-consult--async-search-dwim
+    (&key action type target candidates &allow-other-keys)
+  "DWIM when using a Consult async search command as an ACTION.
+If the TYPE of the target(s) has a notion of associated
+file (files, buffers, libraries and some bookmarks do), then run
+the ACTION with `consult-project-function' set to nil, and search
+only the files associated to the TARGET or CANDIDATES.  For other
+types, run the ACTION with TARGET or CANDIDATES as initial input."
+  (if-let ((file-fn (cdr (assq type embark--associated-file-fn-alist))))
+      (let (consult-project-function)
+        (funcall action
+                 (delq nil (mapcar file-fn (or candidates (list target))))))
+    (funcall action nil (or target (string-join candidates " ")))))
 
 (map-keymap
  (lambda (_key cmd)
-   (cl-pushnew #'embark--cd (alist-get cmd embark-around-action-hooks))
-   (cl-pushnew #'embark-consult--prep-async
-               (alist-get cmd embark-target-injection-hooks)))
+   (unless (eq cmd #'consult-locate)
+     (cl-pushnew cmd embark-multitarget-actions)
+     (cl-pushnew #'embark-consult--async-search-dwim
+                 (alist-get cmd embark-around-action-hooks))))
  embark-consult-async-search-map)
 
 ;;; Tables of contents for buffers: imenu and outline candidate collectors
